from abc import ABC, abstractmethod


class AnswerGenerator(ABC):
    """
    Abstract class for generating answers to questions provided with a context.
    """
    @abstractmethod
    def generate_answer(self, question: str, context: str) -> str:
        pass



class RetrievalQuality:
    def __init__(self, correct: [str], incorrect: [str]):
        self.correct = correct
        self.incorrect = incorrect

    def precision(self):
        return len(self.correct) / (len(self.correct) + len(self.incorrect))

    def total_questions(self):
        return len(self.correct) + len(self.incorrect)


PROVIDER = "bedrock"

MODEL_TITAN_EXPRESS = "amazon.titan-text-express-v1"
EMBEDDING_MODEL_TITAN = "amazon.titan-embed-text-v1"
EMBEDDING_MODEL_TITAN_V2 = "amazon.titan-embed-text-v2:0"

MODEL_ANTHROPIC_SONNET = "anthropic.claude-3-sonnet-20240229-v1:0"
MODEL_ANTHROPIC_HAIKU = "anthropic.claude-3-haiku-20240307-v1:0"
EMBEDDING_MODEL_COHERE_ENG = "cohere.embed-english-v3"
EMBEDDING_MODEL_COHERE_MULTI = "cohere.embed-multilingual-v3"

MODEL_LLAMA_70B = "meta.llama3-70b-instruct-v1:0"

DEFAULT_MODEL = MODEL_TITAN_EXPRESS
DEFAULT_EMBEDDING_MODEL = EMBEDDING_MODEL_TITAN



import json
from typing import List

from openai import OpenAI

from rag4p.integrations.openai import DEFAULT_MODEL
from rag4p.rag.generation.knowledge.knowledge import Knowledge
from rag4p.rag.generation.knowledge.knowledge_extractor import KnowledgeExtractor


class OpenaiKnowledgeExtractor(KnowledgeExtractor):
    def __init__(self, openai_api_key: str, openai_model: str = DEFAULT_MODEL):
        self.openai_client = OpenAI(
            api_key=openai_api_key,
        )
        self.openai_model = openai_model

    def extract_knowledge(self, context: str) -> List[Knowledge]:
        prompt = f"""
Task: Extract Knowledge Chunks

Objective: Extract meaningful knowledge chunks from the provided text. Each chunk should be a distinct, self-contained unit of information presented in a subject-description format. It is essential that the information is from the input text. Do not make any assumptions or add any external information.

Instructions:
1. Identify distinct, relevant pieces of information from the text.
2. Ensure each piece of information focuses on one specific aspect: a person, an event, a location, an activity, a product, a concept, or a term.
3. Consolidate related pieces of information into broader categories only if they contribute to a clearer understanding of the subject.
4. For each piece of information, extract it as a "subject" and provide a corresponding detailed "description" taken from the input text. Do not include your own interpretation or additional information.
5. Ensure that the extracted chunks are formatted as a JSON object or array.
6. Provide enough context in the description to make each knowledge chunk understandable on its own.

Format:
{{
    "knowledge_chunks": [
        {{"subject": "subject", "description": "description"}},
        ...
    ]
}}

Text:
{context}
"""
        completion = self.openai_client.chat.completions.create(
            model=self.openai_model,
            response_format={"type": "json_object"},
            messages=[
                {
                    "role": "system",
                    "content": "You are an assistant that takes apart a piece of text into semantic chunks to be used "
                               "in a RAG system."},
                {"role": "user", "content": prompt},
            ],
            stream=False,
        )

        answer = json.loads(completion.choices[0].message.content)

        k_items = []
        for kc in answer["knowledge_chunks"]:
            item = Knowledge(kc["subject"], kc["description"])
            k_items.append(item)

        return k_items



from weaviate.collections import Collection
import weaviate.classes as wvc
from weaviate.collections.classes.filters import Filter

from rag4p.integrations.weaviate import COLLECTION_NAME
from rag4p.integrations.weaviate.access_weaviate import AccessWeaviate
from rag4p.rag.model.chunk import Chunk
from rag4p.rag.model.relevant_chunk import RelevantChunk
from rag4p.rag.embedding.embedder import Embedder
from rag4p.rag.retrieval.retriever import Retriever


class WeaviateRetriever(Retriever):

    def __init__(self, weaviate_access: AccessWeaviate, embedder: Embedder, additional_properties=None,
                 hybrid: bool = False, collection_name: str = COLLECTION_NAME):
        if additional_properties is None:
            additional_properties = []

        self.weaviate_access = weaviate_access
        self.embedder = embedder
        self.additional_properties = additional_properties
        self.hybrid = hybrid
        self.collection_name = collection_name

    def find_relevant_chunks(self, question: str, max_results: int = 4) -> [RelevantChunk]:
        vector = self.embedder.embed(question)
        if self.hybrid:
            result = self.__chunk_collection().query.hybrid(query=question,
                                                            query_properties=self.additional_properties,
                                                            limit=max_results,
                                                            alpha=0.5,
                                                            fusion_type=wvc.query.HybridFusion.RELATIVE_SCORE,
                                                            vector=vector,
                                                            return_metadata=wvc.query.MetadataQuery(
                                                                distance=True, score=True)
                                                            )
        else:
            result = self.__chunk_collection().query.near_vector(near_vector=vector,
                                                                 limit=max_results,
                                                                 return_metadata=wvc.query.MetadataQuery(distance=True))

        relevant_chunks = []
        for chunk in result.objects:
            properties = {}
            for key in self.additional_properties:
                properties[key] = chunk.properties[key]

            score = chunk.metadata.score if self.hybrid else chunk.metadata.distance
            relevant_chunks.append(RelevantChunk(
                document_id=chunk.properties["documentId"],
                chunk_id=chunk.properties["chunkId"],
                text=chunk.properties["text"],
                total_chunks=chunk.properties["totalChunks"],
                properties=properties,
                score=score,
            ))
        return relevant_chunks

    def get_chunk_by_id(self, document_id: str) -> Chunk:
        parts = document_id.split('_')
        return self.get_chunk(parts[0], int(parts[1]))

    def get_chunk(self, document_id: str, chunk_id: int) -> Chunk:
        filters = Filter.by_property("documentId").equal(document_id) & Filter.by_property("chunkId").equal(chunk_id)

        chunk = self.__chunk_collection().query.fetch_objects(limit=10, filters=filters)
        if len(chunk.objects) == 0:
            raise Exception(f"Chunk with documentId {document_id} and chunkId {chunk_id} not found")
        chunk = chunk.objects[0]

        properties = {}
        for key in self.additional_properties:
            properties[key] = chunk.properties[key]

        return Chunk(
            document_id=chunk.properties["documentId"],
            chunk_id=chunk.properties["chunkId"],
            chunk_text=chunk.properties["text"],
            total_chunks=chunk.properties["totalChunks"],
            properties=properties,
        )

    def loop_over_chunks(self):
        for chunk in self.__chunk_collection().iterator():
            yield from self.__extract_chunk(chunk)

    def __extract_chunk(self, chunk):
        properties = {}
        for key in self.additional_properties:
            properties[key] = chunk.properties[key]

        yield Chunk(
            document_id=chunk.properties["documentId"],
            chunk_id=chunk.properties["chunkId"],
            chunk_text=chunk.properties["text"],
            total_chunks=chunk.properties["totalChunks"],
            properties=properties
        )

    def __chunk_collection(self) -> Collection:
        return self.weaviate_access.client.collections.get(self.collection_name)



from rag4p.integrations.openai.openai_embedder import OpenAIEmbedder
from rag4p.integrations.weaviate import chunk_collection, COLLECTION_NAME
from rag4p.integrations.weaviate.access_weaviate import AccessWeaviate
from rag4p.integrations.weaviate.weaviate_content_store import WeaviateContentStore
from rag4p.indexing.splitters.sentence_splitter import SentenceSplitter
from rag4p.indexing.indexing_service import IndexingService
from rag4p.util.key_loader import KeyLoader
from rag4p.vasa_content_reader import VasaContentReader

import weaviate.classes.config as wvc

if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()

    key_loader = KeyLoader()
    access_weaviate = AccessWeaviate(url=key_loader.get_weaviate_url(), access_key=key_loader.get_weaviate_api_key())
    access_weaviate.force_create_collection(collection_name=COLLECTION_NAME,
                                            properties=chunk_collection.weaviate_properties(
                                                additional_properties=[
                                                    wvc.Property(name="title",
                                                                 data_type=wvc.DataType.TEXT,
                                                                 vectorize_property_name=False,
                                                                 skip_vectorization=True),
                                                    wvc.Property(name="timerange",
                                                                 data_type=wvc.DataType.TEXT,
                                                                 vectorize_property_name=False,
                                                                 skip_vectorization=True),
                                                ]
                                            )
                                            )

    embedder = OpenAIEmbedder(api_key=key_loader.get_openai_api_key())
    content_store = WeaviateContentStore(weaviate_access=access_weaviate,embedder=embedder)
    splitter = SentenceSplitter()
    indexing_service = IndexingService(content_store=content_store)

    content_reader = VasaContentReader()
    indexing_service.index_documents(content_reader=content_reader, splitter=splitter)

    access_weaviate.close()



from abc import ABC, abstractmethod
from typing import List

from rag4p.rag.generation.knowledge.knowledge import Knowledge


class KnowledgeExtractor(ABC):
    """
    Abstract class for extracting knowledge from a context.
    """
    @abstractmethod
    def extract_knowledge(self, context: str) -> List[Knowledge]:
        pass



import json
import os
from abc import ABC, abstractmethod

from openai import OpenAI

from rag4p.rag.generation.chat.chat_prompt import ChatPrompt
from rag4p.rag.generation.quality.answer_from_context_quality import AnswerFromContextQuality
from rag4p.rag.generation.quality.answer_quality import AnswerQuality
from rag4p.rag.generation.quality.answer_to_question_quality import AnswerToQuestionQuality
from rag4p.rag.tracker.rag_observer import RAGObserver


class AnswerQualityService(ABC):

    def determine_quality_of_answer(self, rag_observer: RAGObserver) -> AnswerQuality:
        answer_from_context = self.determine_quality_answer_from_context(rag_observer)
        answer_to_question = self.determine_quality_answer_related_to_question(rag_observer)

        return AnswerQuality(answer_to_question, answer_from_context)

    def determine_quality_answer_related_to_question(self, rag_observer: RAGObserver) -> AnswerToQuestionQuality:
        chat_prompt = ChatPrompt(system_message=self.quality_of_answer_to_question_system_prompt(),
                                 user_message=self.quality_of_answer_to_question_user_prompt())

        completion = self.obtain_answer_to_question_quality(chat_prompt, rag_observer)
        quality, reason = self.split_quality_and_reason(completion)
        return AnswerToQuestionQuality(quality, reason)

    @abstractmethod
    def obtain_answer_to_question_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        pass

    def determine_quality_answer_from_context(self, rag_observer: RAGObserver) -> AnswerFromContextQuality:
        chat_prompt = ChatPrompt(system_message=self.quality_of_answer_from_context_system_prompt(),
                                 user_message=self.quality_of_answer_from_context_user_prompt())

        completion = self.obtain_answer_from_context_quality(chat_prompt, rag_observer)
        quality, reason = self.split_quality_and_reason(completion)
        return AnswerFromContextQuality(quality, reason)

    @abstractmethod
    def obtain_answer_from_context_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        pass

    @staticmethod
    def split_quality_and_reason(quality_and_reason: str) -> (int, str):
        try:
            response = json.loads(quality_and_reason)
        except json.JSONDecodeError:
            print(f"Error: Could not decode response from LLM: {quality_and_reason}")
            return 0, "PROBLEM"
        return int(response["score"]), response["reason"]

    @staticmethod
    def quality_of_answer_to_question_system_prompt():
        return """You are a quality assistant verifying retrieval augmented generation systems. Your task is to 
        verify a generated answer against the proposed question. Give the answer a score between 1 and 5 and keep the 
        number as an integer. 5 means the answer contains the answer to the proposed question completely. 1 means 
        there is no match between the answer and the question at all. Keep the reason short as in maximum 2 sentences. 
        The question provided after 'question:'. The answer after 'answer:'. Write your answers in json format of: 
        {{"score": "score", "reason": "reason"}}
        An example: {{"score": 3,  "reason": "The answer is correct but some details are missing."}}"""


    @staticmethod
    def quality_of_answer_from_context_system_prompt():
        return """You are a quality assistant verifying retrieval augmented generation systems. Your task is to verify 
                a generated answer against the provided context. Give the answer a score between 1 and 5 and keep 
                the number as an integer. 5 means the answer contains only facts from the context. 1 means there is 
                not match between the answer and the provided context at all. If the answer contains exact phrases 
                from the context, the score should be lower as well. Keep the reason short as in maximum 2 sentences. 
                The answer provided after 'answer:'. The context after 'context:'. Write your answers in the json 
                format {{"score": "score", "reason": "reason"}}.  
                An example: {{"score": 3, "reason": "The answer is correct but some details are missing."}}"""

    @staticmethod
    def quality_of_answer_to_question_user_prompt():
        return "Question: {question}\nAnswer: {answer}\nResult:\n"

    @staticmethod
    def quality_of_answer_from_context_user_prompt():
        return "Answer: {answer}\nContext: {context}\nResult:\n"



from openai import OpenAI

from rag4p.rag.generation.question_generator import QuestionGenerator


class OpenAIQuestionGenerator(QuestionGenerator):

    def __init__(self, openai_api_key: str, openai_model: str):
        self.openai_client = OpenAI(
            api_key=openai_api_key,
        )
        self.openai_model = openai_model

    def generate_question(self, context: str) -> str:
        """
        Generates a question from a context.

        :param context: The context to generate a question from.
        :return: The generated question.
        """
        completion = self.openai_client.chat.completions.create(
            model=self.openai_model,
            messages=[
                {"role": "system",
                 "content": "You are a content writer reading a text and writing questions that are answered in that "
                            "text. Use the context as provided and nothing else to come up with one question. The "
                            "question should be a question that a person that does not know a lot about the context "
                            "could ask. Do not use names in your question or exact dates. Do not use the exact words "
                            "in the context. Each question must be one sentence only end always end with a '?' "
                            "character. The context is provided after 'context:'. The result should only contain the "
                            "generated question, nothing else."},
                {"role": "user",
                 "content": f"Context: {context}\nGenerated Question:"},
            ],
            stream=False,
        )
        return completion.choices[0].message.content



from rag4p.integrations.opensearch.index_components import ComponentTemplate, ComponentSettings, \
    ComponentDynamicMappings, ComponentMappings, IndexComponent
from rag4p.integrations.opensearch.opensearch_client import OpenSearchClient


class OpenSearchTemplate:

    def __init__(self,
                 client: OpenSearchClient,
                 index_template: ComponentTemplate,
                 component_settings: ComponentSettings,
                 component_dyn_mappings: ComponentDynamicMappings,
                 component_mappings: ComponentMappings):
        self.client = client
        self.index_template = index_template
        self.component_settings = component_settings
        self.component_dyn_mappings = component_dyn_mappings
        self.component_mappings = component_mappings

    def create_update_template(self) -> list:
        """ Check the version of the current template and update if necessary. """
        print("Initialize or update the product template in OpenSearch.")

        return [
            self.__update_component(component=self.component_settings),
            self.__update_component(component=self.component_dyn_mappings),
            self.__update_component(component=self.component_mappings),
            self.__update_index_template(template=self.index_template)
        ]

    def __update_component(self, component: IndexComponent):
        body = component.get_body()
        required_version = component.version
        name = component.name
        component_needs_update = self.__component_needs_update(component_name=name,
                                                               current_version=required_version)
        if component_needs_update:
            self.client.set_component_template(name=name, body=body)
            result = f"Update the component template {name} to version {required_version}."
        else:
            result = f"The version {required_version} of the component template {name} is up-to-date"

        return result

    def __update_index_template(self, template: ComponentTemplate):
        body = template.get_body()
        required_version = template.version
        name = template.name
        template_needs_update = self.__template_needs_update(template_name=name,
                                                             current_version=required_version)

        if template_needs_update:
            self.client.set_index_template(name=name, body=body)
            result = f"Update the template to version {required_version}."
        else:
            result = f"The version {required_version} of the index template is up-to-date"

        return result

    def __template_needs_update(self, template_name: str, current_version):
        """ The template needs to update if there is no template or if the versions do not match """
        if not self.client.does_index_template_exist(name=template_name):
            print(f"The template with name '{template_name}' is not found")
            return True

        response = self.client.get_index_template(name=template_name)
        print(response)

        templates = response['index_templates']
        if len(templates) != 1:
            raise Exception(f"We cannot have matching more than 1 template while looking for {template_name}")

        index_template = templates[0]['index_template']
        return not (index_template.get("version") == current_version)

    def __component_needs_update(self, component_name: str, current_version):
        """ The template needs to update if there is no template or if the versions do not match """
        print(f"Obtain the component template with name '{component_name}'")
        if not self.client.does_component_template_exist(name=component_name):
            print(f"The component template with name '{component_name}' is not found")
            return True

        response = self.client.get_component_template(name=component_name)
        print(response)

        components = response['component_templates']
        if len(components) != 1:
            raise Exception("We cannot have matching more than 1 component while looking for " + component_name)

        component_template = components[0]['component_template']
        return not (component_template.get("version") == current_version)



from typing import List

from nltk.tokenize import sent_tokenize

from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter
from rag4p.rag.model.chunk import Chunk


class SentenceSplitter(Splitter):
    """
    Splits an InputDocument into Chunks of a single sentence.
    """

    def split(self, input_document: InputDocument, parent_chunk: Chunk = None) -> List[Chunk]:
        input_text = input_document.text if parent_chunk is None else parent_chunk.chunk_text
        sentences = sent_tokenize(input_text)
        chunks = []
        for i in range(len(sentences)):
            chunk_id = str(i) if parent_chunk is None else f"{parent_chunk.chunk_id}_{i}"
            chunk = Chunk(input_document.document_id, chunk_id, len(sentences), sentences[i], input_document.properties)
            chunks.append(chunk)
        return chunks

    @staticmethod
    def name() -> str:
        return "SentenceSplitter"



from rag4p.integrations.ollama import MODEL_PHI3
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.rag.generation.chat.chat_prompt import ChatPrompt
from rag4p.rag.generation.quality.answer_quality_service import AnswerQualityService
from rag4p.rag.tracker.rag_observer import RAGObserver


class OllamaAnswerQualityService(AnswerQualityService):

    def __init__(self, access_ollama: AccessOllama, model: str = MODEL_PHI3):
        self.ollama = access_ollama
        self.model = model

    def obtain_answer_from_context_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        prompt_user = chat_prompt.create_user_message(params={
            "context": rag_observer.context,
            "answer": rag_observer.answer
        })
        prompt_system = chat_prompt.create_system_message(params={})
        prompt = prompt_system + prompt_user
        response = self.ollama.generate_answer(prompt=prompt, model=self.model)
        return response

    def obtain_answer_to_question_quality(self, chat_prompt: ChatPrompt, rag_observer: RAGObserver):
        prompt_user = chat_prompt.create_user_message(params={
            "question": rag_observer.question,
            "answer": rag_observer.answer
        })
        prompt_system = chat_prompt.create_system_message(params={})
        prompt = prompt_system + prompt_user
        response = self.ollama.generate_answer(prompt=prompt, model=self.model)
        return response




from typing import List

import requests

from abc import ABC


class AccessOllama(ABC):
    """
    A simple wrapper for Ollama's API. The wrapper is not trying to be complete, just to provide a simple way to access
    the API. YOu can find more information about the API at https://github.com/ollama/ollama/blob/main/docs/api.md
    """

    def __init__(self, host: str = "localhost", port: int = 11434, protocol: str = "http"):
        self.connection = f"{protocol}://{host}:{port}"

    def list_models(self) -> List[str]:
        """
        List the available models on the provided host, usually you local machine.
        :return: List of strings with the available models in the format name (#params, quantization level).
        """

        response = requests.get(f"{self.connection}/api/tags")
        models = []
        if response.status_code == 200:
            json_response = response.json()
            for model in json_response["models"]:
                models.append(f"{model['name']} ( "
                              f"{model['details']['parameter_size']} - "
                              f"{model['details']['quantization_level']} )")
        return models

    def generate_answer(self, prompt: str, model: str) -> str:
        """
        Generate an answer to a prompt using the provided model.
        :param prompt: Complete prompt to send to the model.
        :param model: The model to use to generate the answer, has to be available in you Ollama instance.
        :return:
        """
        response = requests.post(f"{self.connection}/api/generate",
                                 json={
                                     "prompt": prompt,
                                     "model": model,
                                     "format": "json",
                                     "stream": False
                                 })
        if response.status_code == 200:
            return response.json()["response"]
        raise Exception("Error generating answer:" + response.text)

    def generate_embedding(self, text: str, model: str) -> List[float]:
        response = requests.post(f"{self.connection}/api/embeddings",
                                 json={"model": model, "prompt": text})
        if response.status_code == 200:
            return response.json()["embedding"]

        raise Exception("Error generating embedding:" + response.text)



from rag4p.integrations.openai import MODEL_GPT4_TURBO
from rag4p.integrations.openai.openai_question_generator import OpenAIQuestionGenerator
from rag4p.indexing.splitters.sentence_splitter import SentenceSplitter
from rag4p.indexing.indexing_service import IndexingService
from rag4p.rag.embedding.local.onnx_embedder import OnnxEmbedder
from rag4p.rag.generation.question_generator_service import QuestionGeneratorService
from rag4p.rag.store.local.internal_content_store import InternalContentStore
from rag4p.util.key_loader import KeyLoader
from rag4p.vasa_content_reader import VasaContentReader

if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv()

    key_loader = KeyLoader()
    embedder = OnnxEmbedder()
    content_store = InternalContentStore(embedder=embedder)
    indexing_service = IndexingService(content_store=content_store)
    indexing_service.index_documents(content_reader=VasaContentReader(), splitter=SentenceSplitter())

    question_generator = OpenAIQuestionGenerator(openai_api_key=key_loader.get_openai_api_key(),
                                                 openai_model=MODEL_GPT4_TURBO)

    question_generator_service = QuestionGeneratorService(retriever=content_store,
                                                          question_generator=question_generator)

    question_generator_service.generate_question_answer_pairs(file_name="questions_answers.csv")



class Chunk:
    document_id: str
    chunk_id: str
    total_chunks: int
    chunk_text: str
    properties: {}

    def __init__(self, document_id: str, chunk_id: str, total_chunks: int, chunk_text: str, properties: dict):
        self.document_id = document_id
        self.chunk_id = chunk_id
        self.total_chunks = total_chunks
        self.chunk_text = chunk_text
        self.properties = properties

    def get_id(self):
        return self.document_id + "_" + str(self.chunk_id)



from abc import ABC
from typing import List

from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter
from rag4p.rag.model.chunk import Chunk


class SplitterChain(Splitter):

    def __init__(self, splitters: List[Splitter], include_all_chunks: bool = False):
        self.splitters = splitters
        if not self.splitters or len(self.splitters) == 0:
            raise ValueError("At least one splitter must be provided.")
        self.include_all_chunks = include_all_chunks

    def split(self, input_document: InputDocument, parent_chunk: Chunk = None) -> List[Chunk]:
        return self._split_current_splitter(input_document=input_document)

    def _split_current_splitter(self,
                                input_document: InputDocument,
                                splitter_nr: int = 0,
                                parent_chunk: Chunk = None) -> List[Chunk]:
        chunks = []
        current_chunks = self.splitters[splitter_nr].split(input_document=input_document, parent_chunk=parent_chunk)
        for current_chunk in current_chunks:

            # We always add the first and last chunk, or all chunks if include_all_chunks is True
            if splitter_nr + 1 >= len(self.splitters) or self.include_all_chunks or splitter_nr == 0:
                chunks.append(current_chunk)

            # If we are not at the last splitter, we recursively call the next splitter
            if splitter_nr + 1 < len(self.splitters):
                child_chunks = self._split_current_splitter(input_document=input_document,
                                                            splitter_nr=splitter_nr + 1,
                                                            parent_chunk=current_chunk)
                chunks.extend(child_chunks)

        return chunks

    @staticmethod
    def name() -> str:
        return "SplitterChain"



import json
import os
from abc import ABC
from typing import List

import boto3
from dotenv import load_dotenv

from rag4p.integrations.bedrock import EMBEDDING_MODEL_TITAN_V2
from rag4p.util.key_loader import KeyLoader


class AccessBedrock(ABC):

    def __init__(self, region_name: str = 'eu-central-1', profile_name: str = 'default'):
        self.region_name = region_name
        self.profile_name = profile_name

        self.bedrock = boto3.client(service_name='bedrock', region_name=self.region_name)
        self.bedrock_runtime = boto3.client(service_name='bedrock-runtime', region_name=self.region_name)

    def list_models(self):
        response = self.bedrock.list_foundation_models()
        models = []
        for model in response['modelSummaries']:
            models.append({
                "id": model['modelId'],
                "name": model['modelName'],
                "provider": model['providerName'],
                "output_modalities": model['outputModalities'],
            })
        return models

    def generate_answer(self, prompt: str, model: str) -> str:
        """
        Generate an answer to a prompt using the provided model.
        :param prompt: Complete prompt to send to the model.
        :param model: The model to use to generate the answer, has to be available in you Ollama instance.
        :return:
        """
        # Inference parameters to use.
        temperature = 0.5
        top_p = 1

        # Base inference parameters to use.
        inference_config = {"temperature": temperature, "topP": top_p}

        # Additional inference parameters to use.
        additional_model_fields = {}

        messages = [
            {
                "role": "user",
                "content": [{
                    "text": prompt
                }]
            }
        ]
        # Send the message.
        bedrock_response = self.bedrock_runtime.converse(
            modelId=model,
            messages=messages,
            inferenceConfig=inference_config,
            additionalModelRequestFields=additional_model_fields
        )

        # token_usage = bedrock_response['usage']
        # print("Input tokens: %s", token_usage['inputTokens'])
        # print("Output tokens: %s", token_usage['outputTokens'])
        # print("Total tokens: %s", token_usage['totalTokens'])
        # print("Stop reason: %s", bedrock_response['stopReason'])

        return bedrock_response['output']['message']['content'][0]['text']

    def generate_embedding(self, text: str, model: str) -> List[float]:
        accept = "application/json"
        content_type = "application/json"

        if model.startswith("cohere."):
            body = json.dumps({
                "texts": [text],
                "input_type": "search_document",
            })
        else:
            body = json.dumps({
                "inputText": text,
            })

        response = self.bedrock_runtime.invoke_model(
            body=body, modelId=model, accept=accept, contentType=content_type
        )

        response_body = json.loads(response.get('body').read())

        finish_reason = response_body.get("message")
        if finish_reason is not None:
            raise ValueError(f"Embeddings generation error: {finish_reason}")

        if model.startswith("cohere."):
            return response_body["embeddings"][0]
        else:
            return response_body["embedding"]

    @staticmethod
    def init_from_env(key_loader: KeyLoader):
        region_name = key_loader.get_bedrock_region()
        profile_name = key_loader.get_bedrock_profile()
        return AccessBedrock(region_name=region_name, profile_name=profile_name)


if __name__ == "__main__":
    load_dotenv()

    access_bedrock = AccessBedrock.init_from_env(KeyLoader())
    models_ = access_bedrock.list_models()
    print(models_)

    text = "What is the capital of Germany, is it the same as when Germany was devided into East and West?"
    model_id = "anthropic.claude-3-sonnet-20240229-v1:0"
    response_ = access_bedrock.generate_answer(text, model_id)
    print(response_)

    text = "What is the capital of Germany, is it the same as when Germany was devided into East and West?"
    model_id = "cohere.embed-english-v3"
    embedding = access_bedrock.generate_embedding(text, model_id)
    print(f"Embedding length: {len(embedding)}")



from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter
from rag4p.rag.model.chunk import Chunk


class SingleChunkSplitter(Splitter):
    """
    Does not really split the document, it results in one chunk with the whole document.
    """

    def split(self, input_document: InputDocument, parent_chunk: Chunk = None) -> [Chunk]:
        chunk_id = "0" if parent_chunk is None else f"{parent_chunk.chunk_id}_0"
        chunk_text = input_document.text if parent_chunk is None else parent_chunk.chunk_text
        chunk = Chunk(input_document.document_id, chunk_id, 1, chunk_text, input_document.properties)
        return [chunk]

    @staticmethod
    def name() -> str:
        return "SingleChunkSplitter"



from openai import OpenAI

from rag4p.integrations.openai import DEFAULT_MODEL
from rag4p.rag.generation.answer_generator import AnswerGenerator


class OpenaiAnswerGenerator(AnswerGenerator):
    """
    Used to generate answers to questions using the OpenAI API.
    """
    def __init__(self, openai_api_key: str, openai_model: str = DEFAULT_MODEL):
        self.openai_client = OpenAI(
            api_key=openai_api_key,
        )
        self.openai_model = openai_model

    def generate_answer(self, question: str, context: str) -> str:
        completion = self.openai_client.chat.completions.create(
            model=self.openai_model,
            messages=[
                {"role": "system", "content": "You are an assistant answering questions using the context provided. "
                                              "If the context does not contain the answer, you should tell you cannot "
                                              "answer using the context. The question is provided after 'question:'. "
                                              "The context after 'context:'."},
                {"role": "user", "content": f"Context: {context}\nQuestion: {question}\nAnswer:"},
            ],
            stream=False,
        )

        return completion.choices[0].message.content



from rag4p.integrations.ollama import DEFAULT_EMBEDDING_MODEL
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.rag.embedding.embedder import Embedder


class OllamaEmbedder(Embedder):
    """
    Embedder that uses Ollama's API to embed text.
    """

    def __init__(self, access_ollama: AccessOllama, model: str = DEFAULT_EMBEDDING_MODEL):
        self.ollama = access_ollama
        self.embedding_model = model

    def model(self) -> str:
        return self.embedding_model

    def identifier(self) -> str:
        return f"{self.supplier().lower()}-embedder-{self.embedding_model.lower()}"

    def embed(self, text: str) -> [float]:
        return self.ollama.generate_embedding(text, model=self.embedding_model)

    @staticmethod
    def supplier() -> str:
        return "Ollama"



class InputDocument:
    document_id: str
    text: str
    properties: {}

    def __init__(self, document_id, text, properties):
        self.document_id = document_id
        self.text = text
        self.properties = properties



from typing import List

from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter
from rag4p.rag.generation.knowledge.knowledge_extractor import KnowledgeExtractor
from rag4p.rag.model.chunk import Chunk


class SemanticSplitter(Splitter):

    def __init__(self, knowledge_extractor: KnowledgeExtractor):
        self.knowledge_extractor = knowledge_extractor

    def split(self, input_document: InputDocument, parent_chunk: Chunk = None) -> List[Chunk]:
        input_text = input_document.text if parent_chunk is None else parent_chunk.chunk_text
        knowledge_items = self.knowledge_extractor.extract_knowledge(input_text)

        chunks_ = []
        for i, knowledge_item in enumerate(knowledge_items):
            chunk_id = str(i) if parent_chunk is None else f"{parent_chunk.chunk_id}_{i}"
            chunk_ = Chunk(input_document.document_id,
                           chunk_id,
                           len(knowledge_items),
                           f"{knowledge_item.subject}: {knowledge_item.description}",
                           input_document.properties)
            chunks_.append(chunk_)

        return chunks_

    @staticmethod
    def name() -> str:
        return SemanticSplitter.__name__



import csv
import os

from rag4p.rag.retrieval.quality.question_answer_record import QuestionAnswerRecord
from rag4p.rag.retrieval.quality.retrieval_quality import RetrievalQuality
from rag4p.rag.retrieval.retriever import Retriever


def obtain_retrieval_quality(question_answer_records: [QuestionAnswerRecord], retriever: Retriever) -> RetrievalQuality:
    correct = []
    incorrect = []

    for item in question_answer_records:
        question = item.question
        document_id = item.document_id
        chunk_id = item.chunk_id

        # Retrieve the top 10 chunks for this question
        retrieved_chunks = retriever.find_relevant_chunks(question, max_results=1)[0]

        if retrieved_chunks.document_id == document_id and retrieved_chunks.chunk_id == chunk_id:
            correct.append(item.document_id + " " + str(item.chunk_id))
        else:
            incorrect.append(item.document_id + " " + str(item.chunk_id))

    return RetrievalQuality(correct, incorrect)


def read_question_answers_from_file(file_name: str = "../data/questions_answers.csv"):
    directory = os.getcwd()
    file_path = os.path.join(directory, file_name)

    try:
        questions_answers = []
        with open(file_path, 'r', newline='', encoding='utf-8') as file:
            reader = csv.reader(file)
            next(reader)
            for row in reader:
                document_id = row[0]
                chunk_id = row[1]
                chunk_text = row[2]
                question = row[3]

                questions_answers.append(QuestionAnswerRecord(document_id, chunk_id, chunk_text, question))
        return questions_answers
    except IOError as e:
        print("An error occurred while reading from the file.", e)
        raise



import json

from rag4p.integrations.ollama import DEFAULT_MODEL
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.rag.generation.answer_generator import AnswerGenerator


class OllamaAnswerGenerator(AnswerGenerator):

    def __init__(self, access_ollama: AccessOllama, model: str = DEFAULT_MODEL):
        self.ollama = access_ollama
        self.model = model

    def generate_answer(self, question: str, context: str) -> str:
        prompt = f"""
        You are an assistant answering questions using the context provided. If the context does not contain the
        answer, you should tell you cannot answer using the context. Use the json format for the answer.
        {{"answer": "answer"}}
        question: {question}
        context: {context}
        answer:
        """
        answer = self.ollama.generate_answer(prompt=prompt, model=self.model)
        json_answer = json.loads(answer)
        return json_answer["answer"]



import uuid

import weaviate
import weaviate.classes as wvc


class AccessWeaviate:

    def __init__(self, url, access_key, openai_api_key: str = None):
        print(f"Connecting to Weaviate at {url}")
        headers = {}
        if openai_api_key:
            headers = {"X-OpenAI-Api-Key": openai_api_key}

        self.client = weaviate.connect_to_wcs(
            cluster_url=url,
            auth_credentials=weaviate.auth.AuthApiKey(access_key),
            headers=headers
        )

    def available_collections(self):
        collections = self.client.collections.list_all(simple=True)
        return [collection["name"] for collection in collections]

    def does_collection_exist(self, collection_name):
        exists = self.client.collections.exists(collection_name)

        if not exists:
            print(f"Collection {collection_name} does not exist")

        return exists

    def add_document(self, collection_name: str, properties: dict, vector: [float]):
        self.client.collections.get(collection_name).data.insert(
            uuid=uuid.uuid4(),
            properties=properties,
            vector=vector
        )

    def delete_collection(self, collection_name: str):
        self.client.collections.delete(collection_name)

    def create_collection(self, collection_name: str, properties: list, model: str = "text-embedding-3-small"):
        self.client.collections.create(
            name=collection_name,
            properties=properties,
            vectorizer_config=wvc.config.Configure.Vectorizer.text2vec_openai(
                model=model,
                type_="text",
            )
        )

    def force_create_collection(self, collection_name: str, properties: list):
        if self.does_collection_exist(collection_name):
            self.delete_collection(collection_name)

        self.create_collection(collection_name, properties)

    def close(self):
        self.client.close()

    def print_meta(self):
        meta = self.client.get_meta()
        print(f"Version: {meta['version']}")
        collections = self.client.collections.list_all(simple=False)
        for collection in collections:
            print(f"Available collection: {collection}")

    def obtain_meta(self):
        meta = self.client.get_meta()
        collections = self.client.collections.list_all(simple=False)
        meta["collections"] = collections
        return meta

    def obtain_meta_for_collection(self, collection_name: str):
        meta = self.client.get_meta()
        if self.does_collection_exist(collection_name):
            meta["collection"] = self.client.collections.export_config(name=collection_name)
        return meta

    def query_collection(self, question: str, collection_name: str, max_results: int = 2):
        print(f"Query collection based on user input {question}")

        if not collection_name or not self.does_collection_exist(collection_name):
            print(f"Collection {collection_name} does not exist")
            raise Exception(f"Collection {collection_name} does not exist")

        collection = self.client.collections.get(name=collection_name)
        return collection.query.hybrid(query=question,
                                       limit=max_results,
                                       alpha=0.5,
                                       fusion_type=wvc.query.HybridFusion.RELATIVE_SCORE,
                                       return_metadata=wvc.query.MetadataQuery(
                                           distance=True, score=True)
                                       )




class ChatPrompt:
    def __init__(self,
                 system_message: str = None,
                 user_message: str = None,
                 system_message_filename: str = None,
                 user_message_filename: str = None):
        self.system_message = system_message
        self.user_message = user_message
        if system_message_filename:
            self.system_message = self.__read_from_file(system_message_filename)
        if user_message_filename:
            self.user_message = self.__read_from_file(user_message_filename)

    def create_system_message(self, params: dict):
        return self.system_message.format(**params)

    def create_user_message(self, params: dict):
        return self.user_message.format(**params)

    def __read_from_file(self, filename: str):
        with open(filename, 'r') as file:
            return file.read()


from typing import List

import tiktoken
from tokenizers import Tokenizer

from rag4p.integrations.ollama.ollama_tokenizer import tokenizer_for_model
from rag4p.rag.model.chunk import Chunk
from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter
from rag4p.integrations.openai import DEFAULT_EMBEDDING_MODEL, PROVIDER as OPENAI_PROVIDER
from rag4p.integrations.ollama import PROVIDER as OLLAMA_PROVIDER
from rag4p.integrations.bedrock import PROVIDER as BEDROCK_PROVIDER


class MaxTokenSplitter(Splitter):
    """
    Splits an InputDocument into Chunks of a maximum number of tokens. The tokens are obtained by
    encoding the text of the document using the default model from openai encoding. The chunks of tokens are
    decoded back into text.
    """

    def __init__(self, max_tokens: int = 200, provider: str = OPENAI_PROVIDER, model: str = DEFAULT_EMBEDDING_MODEL):
        self.max_tokens = max_tokens
        self.provider = provider
        if provider == OPENAI_PROVIDER:
            self.encoding = tiktoken.encoding_for_model(model)
        elif provider == OLLAMA_PROVIDER:
            tokenize_model = tokenizer_for_model(model)
            self.encoding = Tokenizer.from_pretrained(tokenize_model)
        elif provider == BEDROCK_PROVIDER:
            # TODO: Implement Bedrock tokenizer, but no information found for the moment, so abuse OpenAI tokenizer
            self.encoding = tiktoken.encoding_for_model(DEFAULT_EMBEDDING_MODEL)
        else:
            raise ValueError(f"Unsupported provider: {provider}")

    def split(self, input_document: InputDocument, parent_chunk: Chunk = None) -> List[Chunk]:
        input_text = input_document.text if parent_chunk is None else parent_chunk.chunk_text
        if self.provider == OPENAI_PROVIDER:
            tokens = self.encoding.encode(input_text)
        elif self.provider == OLLAMA_PROVIDER:
            tokens = self.encoding.encode(input_text).ids
        elif self.provider == BEDROCK_PROVIDER:
            tokens = self.encoding.encode(input_text)
        else:
            raise ValueError(f"Unsupported provider: {self.provider}")

        chunks = []
        num_chunks = len(tokens) // self.max_tokens + (len(tokens) % self.max_tokens != 0)

        while tokens:
            chunk_size = min(len(tokens), self.max_tokens)
            chunk_tokens = tokens[:chunk_size]
            tokens = tokens[chunk_size:]
            chunk_text = self.encoding.decode(chunk_tokens)
            chunk_id = str(len(chunks)) if parent_chunk is None else f"{parent_chunk.chunk_id}_{len(chunks)}"
            chunk = Chunk(input_document.document_id, chunk_id, num_chunks, chunk_text, input_document.properties)
            chunks.append(chunk)

        return chunks

    @staticmethod
    def name() -> str:
        return "MaxTokenSplitter"



import numpy as np

from rag4p.indexing.indexing_service import IndexingService
from rag4p.indexing.splitters.semantic_splitter import SemanticSplitter
from rag4p.indexing.splitters.sentence_splitter import SentenceSplitter
from rag4p.integrations.ollama import MODEL_MISTRAL
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.integrations.ollama.ollama_answer_generator import OllamaAnswerGenerator
from rag4p.integrations.ollama.ollama_embedder import OllamaEmbedder
from rag4p.integrations.ollama.ollama_knowledge_extractor import OllamaKnowledgeExtractor
from rag4p.integrations.ollama.quality.ollama_answer_quality_service import OllamaAnswerQualityService
from rag4p.integrations.openai.openai_knowledge_extractor import OpenaiKnowledgeExtractor
from rag4p.integrations.openai.quality.openai_answer_quality_service import OpenAIAnswerQualityService
from rag4p.rag.generation.observed_answer_generator import ObservedAnswerGenerator
from rag4p.rag.retrieval.strategies.window_retrieval_strategy import WindowRetrievalStrategy
from rag4p.rag.store.local.internal_content_store import InternalContentStore
from rag4p.rag.tracker.rag_tracker import global_data
from rag4p.util.key_loader import KeyLoader
from rag4p.vasa_content_reader import VasaContentReader

if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()

    key_loader = KeyLoader()

    # Using the internal content store with the local embedder
    access_ollama = AccessOllama()
    ollama_model = MODEL_MISTRAL
    embedder = OllamaEmbedder(access_ollama=access_ollama)
    content_store = InternalContentStore(embedder=embedder)
    indexing_service = IndexingService(content_store=content_store)
    splitter = SemanticSplitter(knowledge_extractor=OllamaKnowledgeExtractor(access_ollama=access_ollama, model=ollama_model))
    # splitter = SemanticSplitter(knowledge_extractor=OpenaiKnowledgeExtractor(openai_api_key=key_loader.get_openai_api_key()))
    indexing_service.index_documents(content_reader=VasaContentReader(), splitter=splitter)

    ollama_answer_generator = OllamaAnswerGenerator(access_ollama=access_ollama, model=ollama_model)
    answer_generator = ObservedAnswerGenerator(answer_generator=ollama_answer_generator)

    example_sentences = [
        "How many bolts were replaced?",
        "Since When could people visit the Vasa?",
        "Since when was the Vasa available for the public to visit?",
        "Who was shipwright of the Vasa ship?",
        "Where did the person responsible for building the Vasa ship come from?"
    ]

    # strategy = TopNRetrievalStrategy(retriever=retriever)
    strategy = WindowRetrievalStrategy(retriever=content_store, window_size=1)
    # answer_quality_service = OpenAIAnswerQualityService(openai_api_key=key_loader.get_openai_api_key())
    answer_quality_service = OllamaAnswerQualityService(access_ollama=access_ollama)

    answer_question_quality = []
    answer_context_quality = []
    for example_sentence in example_sentences:
        print("\n----------------------------------")
        retrieval_output = strategy.retrieve_max_results(question=example_sentence, max_results=1)

        # Generate answer using the obtained context
        answer = answer_generator.generate_answer(question=example_sentence,
                                                  context=retrieval_output.construct_context())

        print(f"Question: {example_sentence}")
        print(f"Answer: {answer}")
        print(f"Context: {retrieval_output.construct_context()}")

        rag_observer = global_data["observer"]

        quality = answer_quality_service.determine_quality_answer_related_to_question(rag_observer=rag_observer)
        answer_question_quality.append(quality.quality)
        print(f"Answer to Question Quality: {quality.quality}, Reason: {quality.reason}")

        quality = answer_quality_service.determine_quality_answer_from_context(rag_observer=rag_observer)
        answer_context_quality.append(quality.quality)
        print(f"Answer from Context Quality: {quality.quality}, Reason: {quality.reason}")

        rag_observer.reset()

    print("\n----------------------------------")
    print(f"Question Quality: {np.mean(answer_question_quality)}")
    print(f"Context Quality: {np.mean(answer_context_quality)}")



from abc import ABC


class IndexingResponse(ABC):
    def __init__(self, num_documents: int, num_chunks: int, content_reader: str, splitter: str, running_time: float):
        self.num_documents = num_documents
        self.num_chunks = num_chunks
        self.content_reader = content_reader
        self.splitter = splitter
        self.running_time = running_time

    def __str__(self):
        return (f"IndexingResponse(num_documents={self.num_documents}, "
                f"num_chunks={self.num_chunks}, "
                f"content_reader={self.content_reader}, "
                f"splitter={self.splitter}, "
                f"running_time={self.running_time:.2f} sec.)")



import json
import re
from typing import List

from rag4p.integrations.ollama import DEFAULT_MODEL
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.rag.generation.knowledge.knowledge import Knowledge
from rag4p.rag.generation.knowledge.knowledge_extractor import KnowledgeExtractor


class OllamaKnowledgeExtractor(KnowledgeExtractor):
    def __init__(self, access_ollama: AccessOllama, model: str = DEFAULT_MODEL):
        self.ollama = access_ollama
        self.model = model

    def extract_knowledge(self, context: str) -> List[Knowledge]:
        prompt = f"""
        Task: Extract Knowledge Chunks

        Objective: Extract knowledge chunks from the provided text. Each chunk should be a distinct, self-contained unit of information presented in a subject-description format.

        Instructions:
        1. Identify key pieces of information from the text.
        2. Consolidate related pieces of information into broader categories where possible.
        3. For each consolidated piece of information, extract it as a "subject" and provide a corresponding "description."
        4. Ensure that the extracted chunks are formatted as a JSON object or array.
        5. Typical subjects include: people, places, events, concepts, terms. Broaden the subjects to avoid overly narrow categories.

        Format:
        {{
            "knowledge_chunks": [
                {{"subject": "subject", "description": "description"}},
                ...
            ]
        }}

        Text:
        {context}
        """
        response = self.ollama.generate_answer(prompt=prompt, model=self.model)

        try:
            answer = json.loads(response)
        except json.JSONDecodeError:
            print(f"Error: Could not decode response from Ollama: {response}")
            return []

        k_items = []
        for kc in answer["knowledge_chunks"]:
            if "subject" not in kc or "description" not in kc:
                print(f"Error: Knowledge chunk missing subject or description: {kc}")
                continue
            item = Knowledge(kc["subject"], kc["description"])
            k_items.append(item)

        return k_items


def clean_json_string(s):
    # Trim leading and trailing whitespace
    s = s.strip()
    # Escape control characters
    s = s.encode('unicode_escape').decode('utf-8')
    # Remove trailing commas from arrays and objects
    s = re.sub(r',\s*}', '}', s)
    s = re.sub(r',\s*]', ']', s)
    return s



from rag4p.integrations.ollama import DEFAULT_MODEL
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.integrations.ollama.ollama_answer_generator import OllamaAnswerGenerator
from rag4p.rag.generation.question_generator import QuestionGenerator


class OllamaQuestionGenerator(QuestionGenerator):

    def __init__(self, access_ollama: AccessOllama, model: str = DEFAULT_MODEL):
        self.ollama = access_ollama
        self.model = model

    def generate_question(self, context: str) -> str:
        """
        Generates a question from a context.

        :param context: The context to generate a question from.
        :return: The generated question.
        """
        prompt = f"""
        You are a content writer reading a text and writing questions that are answered in that text. Use the context as
        provided and nothing else to come up with one question. The question should be a question that a person that
        does not know a lot about the context could ask. Do not use names in your question or exact dates. Do not use
        the exact words in the context. Each question must be one sentence only end always end with a '?' character. The
        context is provided after 'context:'. The result should only contain the generated question, nothing else.
        context: {context}
        generated question:
        """
        return self.ollama.generate_answer(prompt=prompt, model=self.model)



import logging
from datetime import datetime

from opensearchpy import OpenSearch, RequestsHttpConnection

search_log = logging.getLogger("search")


class OpenSearchClient:
    def __init__(self, client: OpenSearch, alias_name: str = None):
        self.opensearch = client

        if alias_name:
            self.default_alias_name = alias_name

    def ping(self):
        if self.opensearch.ping():
            print('Connected to OpenSearch')
            return True
        else:
            print('Could not connect to OpenSearch')
            return False

    def create_index(self, provided_alias_name: str = None):
        """
        Create a new index. Name of the index is a combination of the provided or default alias name and a time stamp
        in the format of YearMonthDayHourMinuteSecond. Before the index is created, we remove it if it already exists.
        The settings and mappings are obtained from the shoes_index.json in the config folder.
        :return: The name of the created index
        """
        alias_name = self.__get_alias_name(provided_alias_name)
        index_name = f'{alias_name}-{datetime.now().strftime("%Y%m%d%H%M%S")}'

        self.opensearch.indices.delete(index=index_name, ignore_unavailable=True)
        self.opensearch.indices.create(index=index_name)

        print(f'Created a new index with the name {index_name}')
        return index_name

    def switch_alias_to(self, index_name: str, provided_alias_name: str = None):
        """
        Checks if the alias as configured is already available, if so, remove all indexes it points to. When finished add
        the provided index to the alias.
        :param provided_alias_name: Overrides the default_alias_name.
        :param index_name: Name of the index to assign to the alias
        :return:
        """
        alias_name = self.__get_alias_name(provided_alias_name)
        print(f'Assign alias {alias_name} to {index_name}')
        body = {
            "actions": [
                {"remove": {"index": f'{alias_name}-*', "alias": alias_name}},
                {"add": {"index": index_name, "alias": alias_name}}
            ]
        }
        self.opensearch.indices.update_aliases(body=body)

    def index_document(self, id: str, document: dict, index_name: str):
        """
        Send the provided shoe to Elasticsearch to index that shoe into the provided index.
        :param id: Identifier to use for the index
        :param document: The document to index
        :param index_name: The index to use for indexing the shoe
        :return:
        """
        print(f'Indexing item: {id} into index with name {index_name}')
        self.opensearch.index(index=index_name, id=id, body=document)

    def search(self, body, explain: bool = False, size: int = 10, index_name: str = None):
        index_or_alias = self.__get_alias_name(index_name)
        search_results = self.opensearch.search(index=index_or_alias, body=body, explain=explain, size=size)
        return search_results

    def count_docs(self, index_name: str = None):
        req_index = index_name if index_name is not None else self.default_alias_name
        return self.opensearch.count(index=req_index)

    def set_component_template(self, name, body):
        self.opensearch.cluster.put_component_template(name=name, body=body)

    def set_index_template(self, name, body):
        self.opensearch.indices.put_index_template(name=name, body=body)

    def does_index_template_exist(self, name: str):
        return self.opensearch.indices.exists_index_template(name=name)

    def get_index_template(self, name: str):
        return self.opensearch.indices.get_index_template(name=name)

    def does_component_template_exist(self, name: str):
        return self.opensearch.cluster.exists_component_template(name=name)

    def get_component_template(self, name: str):
        return self.opensearch.cluster.get_component_template(name=name)

    def delete_index(self, index_name: str):
        self.opensearch.indices.delete(index=index_name, ignore_unavailable=True)

    def client(self) -> OpenSearch:
        return self.opensearch

    def __get_alias_name(self, provided_alias_name: str = None) -> str:
        alias_name = provided_alias_name if provided_alias_name is not None else self.default_alias_name

        if not alias_name:
            print("We mandate using aliases for an index. Provided the alias while construction the client"
                            "or provided it with the appropriate function call")
            raise ValueError("We mandate using aliases for an index. Provided the alias while construction the client"
                             "or provided it with the appropriate function call")

        return alias_name



from rag4p.integrations.bedrock.access_bedrock import AccessBedrock
from rag4p.integrations.bedrock.bedrock_answer_generator import BedrockAnswerGenerator
from rag4p.rag.generation.observed_answer_generator import ObservedAnswerGenerator
from rag4p.rag.tracker.rag_tracker import global_data
from rag4p.util.key_loader import KeyLoader


if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()

    key_loader = KeyLoader()

    question = "Since when was the Vasa available for the public to visit?"
    context = ("By Friday 16 February 1962, the ship is ready to be displayed to the general public at the "
               "newly-constructed Wasa Shipyard, where visitors can see Vasa while a team of conservators, "
               "carpenters and other technicians work to preserve the ship.")

    access_bedrock = AccessBedrock()
    bedrock_answer_generator = BedrockAnswerGenerator(access_bedrock=access_bedrock)
    answer_generator = ObservedAnswerGenerator(answer_generator=bedrock_answer_generator)
    answer = answer_generator.generate_answer(question, context)

    print(f"Question: {global_data['observer'].question}")
    print(f"Context: {global_data['observer'].context}")
    print(f"Answer: {answer}")




from abc import abstractmethod, ABC
from typing import List

from rag4p.rag.model.chunk import Chunk
from rag4p.indexing.input_document import InputDocument


class Splitter(ABC):
    """
    Abstract class for splitting an InputDocument or chunks into Chunks. The id and the properties of the document are
    stored in the Chunk with other important data.
    """

    @abstractmethod
    def split(self, input_document: InputDocument, parent_chunk: Chunk = None) -> List[Chunk]:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass
    


from rag4p.rag.retrieval.retrieval_output import RetrievalOutput, RetrievalOutputItem
from rag4p.rag.retrieval.retrieval_strategy import RetrievalStrategy
from rag4p.rag.retrieval.retriever import Retriever


class TopNRetrievalStrategy(RetrievalStrategy):

    def __init__(self, retriever: Retriever):
        self.retriever = retriever

    def retrieve_max_results(self, question: str, max_results: int) -> RetrievalOutput:
        return self.__find_relevant_chunks(question, max_results)

    def retrieve_max_results_observed(self, question: str, max_results: int) -> RetrievalOutput:
        # All observation is done in the ObservedRetriever, use that to observe.
        return self.__find_relevant_chunks(question, max_results)

    def __find_relevant_chunks(self, question: str, max_results: int) -> RetrievalOutput:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)
        retrieval_output_items = []
        for relevant_chunk in relevant_chunks:
            retrieval_output_items.append(RetrievalOutputItem(document_id=relevant_chunk.document_id,
                                                              chunk_id=relevant_chunk.chunk_id,
                                                              text=relevant_chunk.chunk_text))
        return RetrievalOutput(retrieval_output_items)



import base64
import configparser
import os
from configparser import ConfigParser

import requests
from Crypto.Cipher import AES
from Crypto.Util.Padding import unpad, pad


class KeyLoader:
    def __init__(self, key_path: str = None,
                 secret_key: str = None):
        self.key_path = key_path if key_path else "https://www.retrocinevr.nl/rag4j/properties-encrypted.txt"
        self.secret_key = secret_key if secret_key else os.getenv("SECRET_KEY")

    def get_weaviate_api_key(self):
        return self.get_property('WEAVIATE_API_KEY')

    def get_weaviate_url(self):
        return self.get_property('WEAVIATE_URL')

    def get_openai_api_key(self):
        return self.get_property('OPENAI_API_KEY')

    def get_bedrock_region(self):
        return self.get_property('BEDROCK_REGION')

    def get_bedrock_profile(self):
        return self.get_property('BEDROCK_PROFILE')

    def get_property(self, key: str):
        value = os.environ.get(key)
        if not value:
            properties = self.load_properties_from_remote_file(self.key_path)
            value = properties.get(section="DEFAULT", option=key)

        if not value:
            raise Exception(f"Could not find property {key}")

        return value

    def decrypt(self, encrypted_text):
        try:
            encrypted_data = base64.b64decode(encrypted_text)
            cipher = AES.new(self.secret_key.encode(), AES.MODE_ECB)
            decrypted_data = cipher.decrypt(encrypted_data)
            unpadded_data = unpad(decrypted_data, AES.block_size)
            return unpadded_data.decode('utf-8')
        except Exception as e:
            print(f"There was a problem while decrypting the key: {str(e)}")
            raise

    def encrypt(self, str_to_encrypt):
        try:
            cipher = AES.new(self.secret_key.encode(), AES.MODE_ECB)
            padded_data = pad(str_to_encrypt.encode(), AES.block_size)
            encrypted_data = cipher.encrypt(padded_data)
            return base64.b64encode(encrypted_data).decode('utf-8')
        except Exception as e:
            print(f"There was a problem while encrypting the key: {str(e)}")
            raise

    def load_properties_from_remote_file(self, url) -> ConfigParser:
        """
        Load the properties from the remote file and decrypt the values, the file is written using java.util.Properties.
        It has a nasty habit of changing the values, so we need to replace the double backslashes with single ones.
        :param url: The url to load the properties from
        :return: The properties in the format of a ConfigParser
        """
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raise exception if the request failed

            lines = response.text.splitlines()
            lines = [line.replace('\\\\', '\\') for line in lines if not line.startswith('#')]
            properties_content = '\n'.join(lines)

            properties = ConfigParser()
            content = "[DEFAULT]\n" + properties_content
            properties.read_string(content)

            for line in lines:
                key, encrypted_value = line.split('=', 1)
                decrypted_value = self.decrypt(encrypted_value.strip())
                properties.set('DEFAULT', key.strip(), decrypted_value)

            return properties

        except (requests.RequestException, configparser.Error) as e:
            print(f"There was a problem while loading the properties file: {str(e)}")
            raise



from rag4p.integrations.opensearch import DEFAULT_INDEX
from rag4p.integrations.opensearch.opensearch_client import OpenSearchClient
from rag4p.rag.embedding.embedder import Embedder
from rag4p.rag.model.chunk import Chunk
from rag4p.rag.model.relevant_chunk import RelevantChunk
from rag4p.rag.retrieval.retriever import Retriever


class OpenSearchRetriever(Retriever):

    def __init__(self, opensearch_client: OpenSearchClient, embedder: Embedder,
                 additional_properties=None, hybrid: bool = False, index_name: str = DEFAULT_INDEX):
        if additional_properties is None:
            additional_properties = []

        self.opensearch_client = opensearch_client
        self.embedder = embedder
        self.index_name = index_name
        self.additional_properties = additional_properties
        self.hybrid = hybrid

    def find_relevant_chunks(self, question: str, max_results: int = 4) -> [RelevantChunk]:
        fields = ["chunk_text"]
        fields.extend(self.additional_properties)
        if not self.hybrid:
            query = {"match_all": {}}
        else:
            query = {"multi_match": {"fields": fields, "query": question}}
        query = {
            "_source": {
                "excludes": ["chunk_vector"]
            },
            "query": {
                "bool": {
                    "should": [
                        {
                            "script_score": {
                                "query": {"match_all": {}},
                                "script": {
                                    "source": "knn_score",
                                    "lang": "knn",
                                    "params": {
                                        "field": "chunk_vector",
                                        "query_value": self.embedder.embed(question),
                                        "space_type": "cosinesimil"
                                    }
                                }
                            }
                        },
                        query
                    ]
                }
            },
            "size": max_results
        }
        search_response = self.opensearch_client.search(body=query, index_name=self.index_name, size=max_results)
        relevant_chunks = []

        if search_response["hits"]["total"]["value"] == 0:
            return []

        for hit in search_response["hits"]["hits"]:
            properties = {}
            for key in self.additional_properties:
                properties[key] = hit["_source"][key]

            score = hit["_score"]
            relevant_chunks.append(RelevantChunk(
                document_id=hit["_source"]["document_id"],
                chunk_id=hit["_source"]["chunk_id"],
                text=hit["_source"]["chunk_text"],
                total_chunks=hit["_source"]["total_chunks"],
                properties=properties,
                score=score,
            ))
        return relevant_chunks

    def get_chunk_by_id(self, chunk_id: str) -> Chunk:
        chunk_response = self.opensearch_client.client().get(index=self.index_name, id=chunk_id)

        return Chunk(
            document_id=chunk_response["_source"]["document_id"],
            chunk_id=chunk_response["_source"]["chunk_id"],
            chunk_text=chunk_response["_source"]["chunk_text"],
            total_chunks=chunk_response["_source"]["total_chunks"],
            properties=chunk_response["_source"],
        )

    def loop_over_chunks(self):
        raw_client = self.opensearch_client.client()

        scroll_time = "2m"
        page_size = 1000

        response = raw_client.search(index=self.index_name, scroll=scroll_time, size=page_size)
        scroll_id = response['_scroll_id']
        hits = response['hits']['hits']

        chunks = []
        while len(hits) > 0:
            # Process the current batch of documents
            for hit in hits:
                properties = {}
                for key in self.additional_properties:
                    properties[key] = hit["_source"][key]

                chunks.append(Chunk(
                    document_id=hit["_source"]["document_id"],
                    chunk_id=hit["_source"]["chunk_id"],
                    chunk_text=hit["_source"]["chunk_text"],
                    total_chunks=hit["_source"]["total_chunks"],
                    properties=properties,
                ))

            # Fetch the next batch of documents
            response = raw_client.scroll(
                scroll_id=scroll_id,
                scroll=scroll_time
            )
            # Extract the scroll ID and the next batch of documents
            scroll_id = response['_scroll_id']
            hits = response['hits']['hits']

        # Clean up the scroll context when done
        raw_client.clear_scroll(scroll_id=scroll_id)

        return chunks



from rag4p.integrations.openai.openai_embedder import OpenAIEmbedder
from rag4p.integrations.weaviate.access_weaviate import AccessWeaviate
from rag4p.integrations.weaviate.weaviate_retriever import WeaviateRetriever
from rag4p.util.key_loader import KeyLoader

if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv()

    key_loader = KeyLoader()

    client = AccessWeaviate(url=key_loader.get_weaviate_url(), access_key=key_loader.get_weaviate_api_key())

    client.print_meta()

    embedder = OpenAIEmbedder(api_key=key_loader.get_openai_api_key())
    retriever = WeaviateRetriever(weaviate_access=client,
                                  embedder=embedder,
                                  additional_properties=["title", "speakers"],
                                  hybrid=True)

    for chunk in retriever.loop_over_chunks():
        print(f"Document: {chunk.document_id} - {chunk.chunk_id}")

    get_chunk = retriever.get_chunk(document_id="using-ai-to-save-time-and-sanity", chunk_id=0)
    print("--------------------------------------------------")
    print(f"Document: {get_chunk.document_id} - {get_chunk.chunk_id}")

    relevant_chunks = retriever.find_relevant_chunks("What are jettro and daniel talking about?", max_results=2)
    print(f"Found {len(relevant_chunks)} relevant chunks for query: What are jettro and daniel talking about?")

    for chunk in relevant_chunks:
        print(f"Document: {chunk.document_id}")
        print(f"Chunk id: {chunk.chunk_id}")
        print(f"Title: {chunk.properties['title']}")
        print(f"Speakers: {chunk.properties['speakers']}")
        print(f"Text: {chunk.chunk_text}")
        print(f"Score: {chunk.score}")
        print("--------------------------------------------------")

    client.close()



from typing import List


class RetrievalOutputItem:
    document_id: str
    chunk_id: int
    text: str

    def __init__(self, document_id, chunk_id, text):
        self.document_id = document_id
        self.chunk_id = chunk_id
        self.text = text

    pass


class RetrievalOutput:
    items: List[RetrievalOutputItem]

    def __init__(self, items):
        self.items = items

    def construct_context(self):
        return ' '.join(item.text for item in self.items)




class QuestionAnswerRecord:
    def __init__(self, document_id: str, chunk_id: str, chunk_text: str, question: str):
        self.document_id = document_id
        self.chunk_id = chunk_id
        self.chunk_text = chunk_text
        self.question = question



class AnswerToQuestionQuality:
    """
    Represents the quality of an answer to a question. The quality is a number between 1 and 5, where 5
    means that the answer is a complete answer to the question and 1 means that the answer unrelated to the
    question.
    """
    quality: int
    reason: str

    def __init__(self, quality: int, reason: str):
        self.quality = quality
        self.reason = reason



from rag4p.integrations.bedrock import DEFAULT_MODEL
from rag4p.integrations.bedrock.access_bedrock import AccessBedrock
from rag4p.rag.generation.answer_generator import AnswerGenerator


class BedrockAnswerGenerator(AnswerGenerator):
    def __init__(self, access_bedrock: AccessBedrock, model: str = DEFAULT_MODEL):
        self.access_bedrock = access_bedrock
        self.model = model

    def generate_answer(self, question: str, context: str) -> str:
        prompt = f"""
        You are an assistant answering questions using the context provided. If the context does not contain the
        answer, you should tell you cannot answer using the context.
        question: {question}
        context: {context}
        answer:
        """
        return self.access_bedrock.generate_answer(prompt=prompt, model=self.model)




from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter_chain import SplitterChain
from rag4p.indexing.splitters.section_splitter import SectionSplitter
from rag4p.indexing.splitters.semantic_splitter import SemanticSplitter
from rag4p.integrations.ollama import MODEL_GEMMA2, MODEL_PHI3
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.integrations.ollama.ollama_knowledge_extractor import OllamaKnowledgeExtractor
from rag4p.integrations.openai.openai_embedder import OpenAIEmbedder
from rag4p.integrations.openai.openai_knowledge_extractor import OpenaiKnowledgeExtractor
from rag4p.util.key_loader import KeyLoader

input_text = f"""Ever thought about building your very own question-answering system? Like the one that powers Siri, 
Alexa, or Google Assistant? Well, we've got something awesome lined up for you! In our hands-on workshop, we'll guide 
you through the ins and outs of creating a question-answering system. We prefer using Python for the workshop. We 
have prepared a GUI that works with python. If you prefer another language, you can still do the workshop, 
but you will miss the GUI to test your application.

You'll get your hands dirty with vector stores and Large Language Models, we help you combine these two in a way 
you've never done before. You've probably used search engines for keyword-based searches, right? Well, prepare to 
have your mind blown. We'll dive into something called semantic search, which is the next big thing after traditional 
searches. It’s like moving from asking Google to search "best pizza places" to "Where can I find a pizza place that 
my gluten-intolerant, vegan friend would love?" – you get the idea, right?

We’ll be teaching you how to build an entire pipeline, starting from collecting data from various sources, 
converting that into vectors (yeah, it’s more math, but it’s cool, we promise), and storing it so you can use it to 
answer all sorts of queries. It's like building your own mini Google!

We've got a repository ready to help you set up everything you need on your laptop. By the end of our workshop, 
you'll have your question-answering system ready and running. So, why wait? Grab your laptop, bring your coding hat, 
and let's start building something fantastic together. Trust us, it’s going to be a blast!

Some of the highlights of the workshop: 
- Use a vector store (OpenSearch, Elasticsearch, Weaviate)
- Use a Large Language Model (OpenAI, HuggingFace, Cohere, PaLM, Bedrock)
- Use a tool for content extraction (Unstructured, Llama)
- Create your pipeline (Langchain, Custom)
"""

if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv()

    key_loader = KeyLoader()
    embedder = OpenAIEmbedder(api_key=key_loader.get_openai_api_key())
    extractor = OpenaiKnowledgeExtractor(openai_api_key=key_loader.get_openai_api_key())
    # access_ollama = AccessOllama()
    # extractor = OllamaKnowledgeExtractor(access_ollama=access_ollama, model="llama3.1")

    splitter = SplitterChain(splitters=[SectionSplitter(), SemanticSplitter(knowledge_extractor=extractor)],
                             include_all_chunks=True)

    input_document = InputDocument(document_id="doc1", text=input_text, properties={})

    chunks = splitter.split(input_document)

    for chunk in chunks:
        print(f"{chunk.chunk_id} ({chunk.total_chunks}): {chunk.chunk_text}")



from rag4p.indexing.indexing_service import IndexingService
from rag4p.indexing.splitters.sentence_splitter import SentenceSplitter
from rag4p.integrations.ollama.access_ollama import AccessOllama
from rag4p.integrations.ollama.ollama_embedder import OllamaEmbedder
from rag4p.rag.store.local.internal_content_store import InternalContentStore
from rag4p.util.key_loader import KeyLoader
from rag4p.vasa_content_reader import VasaContentReader

if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()

    key_loader = KeyLoader()
    access_ollama = AccessOllama()
    embedder = OllamaEmbedder(access_ollama=access_ollama)
    content_reader = VasaContentReader()
    content_store = InternalContentStore(embedder=embedder)
    splitter = SentenceSplitter()

    indexing_service = IndexingService(content_store=content_store)
    response = indexing_service.index_documents(content_reader=content_reader, splitter=splitter)

    print(response)

    query = "Since when was the Vasa available for the public to visit?"
    relevant_chunks = content_store.find_relevant_chunks(query=query, max_results=2)
    print(f"Found {len(relevant_chunks)} relevant chunks for query: {query}")
    for chunk in relevant_chunks:
        print(f"Document: {chunk.document_id}")
        print(f"Chunk id: {chunk.chunk_id}")
        print(f"Text: {chunk.chunk_text}")



import re
from typing import List

from rag4p.indexing.input_document import InputDocument
from rag4p.indexing.splitter import Splitter
from rag4p.rag.model.chunk import Chunk


class SectionSplitter(Splitter):
    def split(self, input_document: InputDocument, parent_chunk: Chunk = None) -> List[Chunk]:
        input_text = input_document.text if parent_chunk is None else parent_chunk.chunk_text
        sections = re.split(r"\n\s*\n", input_text)

        chunks_ = []
        for i, section in enumerate(sections):
            chunk_id = str(i) if parent_chunk is None else f"{parent_chunk.chunk_id}_{i}"
            chunk_ = Chunk(input_document.document_id, chunk_id, len(sections), section, input_document.properties)
            chunks_.append(chunk_)

        return chunks_

    @staticmethod
    def name() -> str:
        return SectionSplitter.__name__



from rag4p.rag.retrieval.retrieval_strategy import RetrievalStrategy
from rag4p.rag.retrieval.retriever import Retriever
from rag4p.rag.retrieval.retrieval_output import RetrievalOutput, RetrievalOutputItem
from rag4p.rag.model.relevant_chunk import RelevantChunk
from rag4p.rag.tracker.rag_tracker import global_data


class DocumentRetrievalStrategy(RetrievalStrategy):
    """
    This retrieval strategy retrieves complete documents for the relevant chunks. This strategy is useful when you want
    to compare parts of a document to vectors, but want the complete document for the context of an LLM.

    Imagine you have blogs posts as documents and you want to asnwer questions about them. The body of the blog is
    splitted into chunks, now you want to answer who wrote about a specific topic. The authors of the blogs posts are
    in the additional properties of the chunks. You can use this strategy to retrieve the complete blog post and the
    author of the blog post. Using this context, the LLM can find the author of the specific found blog post.

    When the chunks are created using a chain of splitters, the complete document is retrieved by combining all chunks
    from the highest level of the chain.
    """

    def __init__(self, retriever: Retriever, observe: bool = False):
        self.retriever = retriever
        self.observe = observe

    def retrieve_max_results(self, question, max_results) -> RetrievalOutput:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)
        return self.__extract_document_for_chunk(relevant_chunks, self.observe)

    def retrieve_max_results_observed(self, question, max_results) -> RetrievalOutput:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)

        return self.__extract_document_for_chunk(relevant_chunks, observe=True)

    def __extract_document_for_chunk(self, relevant_chunks: [RelevantChunk], observe: bool = False) -> RetrievalOutput:
        retrieval_output_items = []

        # Remove chunks from the same document
        unique_docs = self.__unique_documents_from_chunks(relevant_chunks)

        for relevant_chunk in unique_docs:
            # for a chunk that is not of the first splitter level, we need to obtain the parent chunk that is of the
            # first splitter level
            if len(relevant_chunk.chunk_id.split("_")) > 1:
                relevant_chunk = self.retriever.get_chunk(relevant_chunk.document_id, relevant_chunk.chunk_id.split("_")[0])

            overall_text = self.__read_text_from_all_chunks_for_document(relevant_chunk.document_id,
                                                                         relevant_chunk.total_chunks)

            # A chunk can have additional properties, add them to the text as well in the format of key: value
            for key, value in relevant_chunk.properties.items():
                overall_text += f"\n{key}: {value} "

            relevant_item = RetrievalOutputItem(document_id=relevant_chunk.document_id,
                                                chunk_id=relevant_chunk.chunk_id,
                                                text=overall_text)
            retrieval_output_items.append(relevant_item)

            if observe:
                print(f"'{relevant_chunk.get_id()}'")
                print(f"'{overall_text}'")
                global_data["observer"].add_relevant_chunk(relevant_chunk.get_id(), overall_text)

        return RetrievalOutput(retrieval_output_items)

    def __read_text_from_all_chunks_for_document(self, document_id, total_chunks):
        """
        A document consists of multiple numbered chunks, we need to combine all chunks to get the complete document.
        :param document_id: Document to get all chunks for.
        :param total_chunks: Total amount of chunks in the document.
        :return: The complete text of all chunks combined.
        """
        chunk_ids = range(0, total_chunks)
        overall_text = ""
        for chunk_id in chunk_ids:
            chunk = self.retriever.get_chunk(document_id, chunk_id)
            overall_text += chunk.chunk_text + " "
        return overall_text

    @staticmethod
    def __unique_documents_from_chunks(relevant_chunks):
        """
        Remove chunks from the same document, as we are returning complete documents.
        :param relevant_chunks: Provided relevant chunks
        :return: Relevant chunks from unique documents only
        """
        unique_docs_dict = {chunk.document_id: chunk for chunk in relevant_chunks}
        return list(unique_docs_dict.values())



from rag4p.connectweaviate.access_weaviate import AccessWeaviate
from rag4p.util.key_loader import KeyLoader

if __name__ == '__main__':
    from dotenv import load_dotenv
    load_dotenv()

    key_loader = KeyLoader()
    weaviate_client = AccessWeaviate(url=key_loader.get_weaviate_url(), access_key=key_loader.get_weaviate_api_key())
    try:
        weaviate_client.print_meta()
    finally:
        weaviate_client.close()



from rag4p.integrations.openai.openai_embedder import OpenAIEmbedder
from rag4p.integrations.opensearch.connection_builder import build_aws_search_service
from rag4p.integrations.opensearch.opensearch_client import OpenSearchClient
from rag4p.integrations.opensearch.opensearch_retriever import OpenSearchRetriever
from rag4p.integrations.weaviate.access_weaviate import AccessWeaviate
from rag4p.integrations.weaviate.weaviate_retriever import WeaviateRetriever
from rag4p.util.key_loader import KeyLoader

if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv()
    key_loader = KeyLoader()

    opensearch_conn = build_aws_search_service(stack_name=key_loader.get_property("OPENSEARCH_STACK_NAME"),
                                               application_prefix=key_loader.get_property("OPENSEARCH_APP_PREFIX"))
    client = OpenSearchClient(client=opensearch_conn)

    embedder = OpenAIEmbedder(api_key=key_loader.get_openai_api_key())
    retriever = OpenSearchRetriever(opensearch_client=client,
                                    embedder=embedder,
                                    index_name="rag4p-vasa",
                                    additional_properties=["title", "timerange"],
                                    hybrid=False)

    for chunk in retriever.loop_over_chunks():
        print(f"Document: {chunk.document_id} - {chunk.chunk_id}")

    get_chunk = retriever.get_chunk(document_id="a-faithful-contract", chunk_id=4)
    print("--------------------------------------------------")
    print(f"Document: {get_chunk.document_id} - {get_chunk.chunk_id}")

    question = "shipwright"
    relevant_chunks = retriever.find_relevant_chunks(question=question, max_results=2)
    print(f"Found {len(relevant_chunks)} relevant chunks for query: {question}")

    for chunk in relevant_chunks:
        print(f"Document: {chunk.document_id}")
        print(f"Chunk id: {chunk.chunk_id}")
        print(f"Title: {chunk.properties['title']}")
        print(f"Time range: {chunk.properties['timerange']}")
        print(f"Text: {chunk.chunk_text}")
        print(f"Score: {chunk.score}")
        print("--------------------------------------------------")



from rag4p.rag.model.relevant_chunk import RelevantChunk
from rag4p.rag.retrieval.retrieval_output import RetrievalOutput, RetrievalOutputItem
from rag4p.rag.retrieval.retrieval_strategy import RetrievalStrategy
from rag4p.rag.retrieval.retriever import Retriever
from rag4p.rag.tracker.rag_tracker import global_data


class WindowRetrievalStrategy(RetrievalStrategy):
    """
    This retrieval strategy retrieves a window of chunks around the relevant chunks. This strategy is useful when you
    want to provide more context to the LLM. The window size can be set to 1, 2, 3, etc. The window size is the number
    of chunks before and after the relevant chunk that should be included in the context.

    When multiple levels of chunks are used, the window is applied to the level of chunks that are relevant. This means
    that the window is applied to the level of chunks that are returned by the retriever.
    """

    def __init__(self, retriever: Retriever, window_size: int = 1):
        self.retriever = retriever
        self.window_size = window_size

    def retrieve_max_results(self, question, max_results) -> RetrievalOutput:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)

        return self.__extract_window_for_relevant_chunks(relevant_chunks)

    def retrieve_max_results_observed(self, question, max_results) -> RetrievalOutput:
        relevant_chunks = self.retriever.find_relevant_chunks(question, max_results)

        return self.__extract_window_for_relevant_chunks(relevant_chunks, observe=True)

    def __extract_window_for_relevant_chunks(self, relevant_chunks: [RelevantChunk], observe: bool = False) \
            -> RetrievalOutput:
        retrieval_output_items = []
        for relevant_chunk in relevant_chunks:
            chunk_ids = self.__chunk_ids_for_window(relevant_chunk.chunk_id,
                                                    self.window_size,
                                                    relevant_chunk.total_chunks)
            overall_text = ""
            for chunk_id in chunk_ids:
                chunk = self.retriever.get_chunk(relevant_chunk.document_id, chunk_id)
                overall_text += chunk.chunk_text + " "

            if observe:
                global_data["observer"].add_relevant_chunk(relevant_chunk.get_id(), overall_text)

            relevant_item = RetrievalOutputItem(document_id=relevant_chunk.document_id,
                                                chunk_id=relevant_chunk.chunk_id,
                                                text=overall_text)
            retrieval_output_items.append(relevant_item)
        return RetrievalOutput(retrieval_output_items)

    @staticmethod
    def __chunk_ids_for_window(chunk_id: str, window_size: int, number_of_chunks: int) -> [int]:
        # The chunk_id has the format 0 or 0_0, we need to extract the last part
        _chunk_id = int(chunk_id.split("_")[-1])

        # Calculate the start and end of the window
        start = max(0, _chunk_id - window_size)
        end = min(number_of_chunks - 1, _chunk_id + window_size)

        # Generate the list of chunk_ids
        chunk_ids = [str(start + i) for i in range(end - start + 1)]

        return chunk_ids
